#ifndef __AAP_H__
#define __AAP_H__

#include "tim.h"
#include "net.h"
#include "net_socket.h"
#include "cfg.h"
#include "addr_range.h"
#include "arec.h"

/**
 * aap.h
 *
 * The aap module is the public interface to AAP. The interface
 * consists of the function aap_new() and the methods start() and
 * allocate(). To use the aap module, first create net, tim and cfg
 * objects, the call aap_new() with there objects. The AAP server
 * is started when the start() method is called. To allocate
 * addresses, call the allocate() method. When the requested number
 * of addresses are allocated, the callback function f in invoked
 * with the void *obj as its first parameter, and an addr_range *ar
 * as the second. The addr_range *ar will contain the address range
 * that has been allocated by aap.
 *
 * The following example illustrates a typical scenario, by allocating
 * 20 multicast addresses with a lifetime of 10 minutes:
 *
 * static void claimed(void *obj, addr_range *ar) {
 *    char *msg = obj;
 *    printf("%s: allocated an address range\n", msg);
 * }
 *
 * int main(int argc, char **argv) {
 *    net *n;
 *    tim *t;
 *    cfg *c;
 *    aap *a;
 *
 *    t = tim_new();
 *    n = net_new(t);
 *    c = somehow_create_a_cfg_object();
 *
 *    a = aap_new(n, t, c);
 *    a->start(a);
 *    a->allocate(a, 20, time(NULL) + 600, claimed, "Hey!");
 *
 *    n->main(n);
 * }
 * 
 * 
 */
typedef struct aap_ aap;

enum aap_state {
    AAP_STARTUP,
    AAP_RUNNING
};
  
typedef void (* aap_callback_func)(void *obj, addr_range *ar);



struct aap_ {

    net *net;
    net_socket *in_ns, *out_ns;
    tim *tim;
    list *claim_list;
    list *defens_list;
    cfg *cfg;

    tim_id_t asa_exp_timer;
    tim_id_t asa_resend_timer;
    int last_asa_exp_time;
    int last_asa_receive_time;
    char *last_asa_body;
    int last_asa_body_len;

    int asrp_timer;
    int rseq;
    enum aap_state state;
    list *asa_defenders;
    arec *arec;

    unsigned int startup_time;

    void (*allocate)(aap *aap, int numallocs, unsigned int end_time,
                     aap_callback_func f, void *obj);

    void (*start)(aap *aap);
    int (* next_rseq)(aap *aap);

};

aap *aap_new(net *, tim *, cfg *);

#endif /* __AAP_H__ */
