#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "aap_aclm.h"
#include "aap_aiu.h"

#include "defender.h"
#include "claimer.h"

/*------------------------------------------------------------------*/
/* The ACLM message header */
struct aclm_raw {
    unsigned int current_time;
    addr_range_raw ranges[500];
};

  
/*------------------------------------------------------------------*/
/*
 * send_it format and sends an ACLM message.
 *
 */
static int send_msg(aap_aclm *a) {
    aap_msg *msg;
    struct aclm_raw raw_aclm;
    addr_range *ar;
    int len;
    /* construct the message body */
    bzero(&raw_aclm, sizeof(struct aclm_raw));

    /* insert current time */
    raw_aclm.current_time = htonl(time(NULL));

    /* copy the raw address range into claim */
    ar = a->alloc->range(a->alloc);
    raw_aclm.ranges[0] = ar->raw(ar);

    /* len is the length of the message body */
    len = 4 + sizeof(addr_range_raw);

    msg = aap_msg_new();
    msg->create(msg, AAP_MSG_ACLM, AAP_MSG_IP,
                a->rseq, a->mseq, (char *)&raw_aclm, len);
    msg->send(msg, a->n, a->ns);
    a->mseq++;
    return 0;
}
/*------------------------------------------------------------------*/
/*
 * delete deletes the aclm object.
 *
 */
static void delete(aap_aclm *a) {
    /*  a->alloc->delete(a->alloc);*/
    free(a);
}
/*------------------------------------------------------------------*/
static addr_range *range(aap_aclm *a) {
    return a->alloc->range(a->alloc);
}
/*------------------------------------------------------------------*/
/*
 * aap_aclm_receive is called when an ACLM message in received. It
 * computes the clock skew, and checks whether the claim is feasible.
 * If not, it takes the appropriate measures, which is to send an AIU
 * if the claim collided with one of our allocations, or sets a timer
 * in case the collision was with another MAAS' allocation.
 *
 */

int aap_aclm_receive(aap *aap, aap_msg *msg) {
  
    int len, i, j, k, pos, defending;
    struct aclm_raw *aclm;
    unsigned int clock_skew;
    int rseq, mseq, addrfamily;
    list *overlaps, *l;
    addr_range *ar;
    allocation *tmp_alloc;
  
    rseq = msg->rseq(msg);
    mseq = msg->mseq(msg);
    addrfamily = msg->addrfamily(msg);
  
    /* extract the message body into aclm, and length into len */
    msg->body(msg, (char **)&aclm, &len);
  
    /* compute the claim time, adjust for clock skew
       as in section 5.3.1 of draft-ietf-malloc-aap-03.txt */
    clock_skew = ntohl(aclm->current_time) - time(NULL);
  
    /* i contains the number of ranges */
    i = (len - 4) / sizeof(addr_range_raw);
    if(len != (i * sizeof(addr_range_raw) + 4)) {
        /* malformed message */
        DEBUG('d', "aap_aclm: Received malformed message.\n");
        return 1;
    }
  
    /* process the claimed address ranges */
    while(i > 0) {

    
        ar = addr_range_new_from_raw(&(aclm->ranges[i - 1]));
        /* adjust for clock skew */
        ar->adjust(ar, clock_skew);

        /* check to see if the incomming ACLM is a reclaim for an
           address range that we are defending. if so, the
           defens should be canceled */
        l = list_new();
        len = aap->defens_list->length(aap->defens_list);
        for(j = 0; j < len; j++) {
            defender *def = aap->defens_list->elem_at(aap->defens_list, j);
            if(def->match(def, rseq, msg->sender(msg))) {
                l->insert(l, def);
            }
        }

        len = l->length(l);
        for(j = 0; j < len; j++) {
            defender *def = l->elem_at(l, j);
            def->cancel(def);
            pos = aap->defens_list->find(aap->defens_list, def);
            aap->defens_list->remove(aap->defens_list, pos);
            def->delete(&def);
        }
        l->delete(l);
        l = NULL;

        /* check if the incomming ACLM overlaps an allocation
           in the allocation record, in which case it should be
           defended. */


        defending = 0;
        for(k = 0; k < aap->defens_list->length(aap->defens_list); k++) {
            defender *def = aap->defens_list->elem_at(aap->defens_list, k);
            if(def->overlap(def, ar)) {
                defending++;
            }
      
        }

        if(!defending) {
            overlaps = aap->arec->overlap(aap->arec, ar);
            if(!overlaps->is_empty(overlaps)) {      
                for( j = 0 ; j < overlaps->length(overlaps) ; j++) {
                    tmp_alloc = overlaps->elem_at(overlaps, j);
	  
                    if((tmp_alloc->owner_addr(tmp_alloc) != msg->sender(msg)) &&
                        (tmp_alloc->flag != ALLOCATION_PREALLOC)) {
                        /* Above added by Fredrik. Don't defend preallocations. */
                        defender *def = defender_new(aap, tmp_alloc, rseq, msg->sender(msg));	  
                        def->defend(def);
                        def->aap->defens_list->insert(def->aap->defens_list, def);
                    }
                }
            }
            overlaps->delete(overlaps);
            overlaps = NULL;
        }
        /* check if the incomming ACLM overlaps with any outstanding claims
           of ours, in which case the claim should be canceled. */

        overlaps = aap->arec->overlap_claims(aap->arec, ar);
        if(overlaps->is_empty(overlaps)) {
            aap->arec->insert_claim(aap->arec, ar);
        }

	
        len = aap->claim_list->length(aap->claim_list);
        for(j = 0; j < len; j++) {

            claimer *clm = aap->claim_list->elem_at(aap->claim_list, j);

            if(!aap->net->is_self(aap->net, msg->sender(msg)) &&
               clm->is_active(clm) &&
               ar->overlap(ar, clm->range(clm))) {
                int retry_wait = 5 + ((double)(rand())/RAND_MAX) * 5; 

                printf("Overlapping ACLM received!\n");
                /* The overlap has already been inserted */
                clm->retry_in(clm, retry_wait);
            }
        }

        if(!overlaps->is_empty(overlaps)) {
            ar->delete(ar);
            ar = NULL;
        }
        overlaps->delete(overlaps);
        overlaps = NULL;
        i--;    
    
    }
    return 0;
}


/*------------------------------------------------------------------*/
static void set_allocation(aap_aclm *aclm, allocation *alloc) {
    aclm->alloc = alloc;
}

/*------------------------------------------------------------------*/
aap_aclm *aap_aclm_new(net *n, net_socket *ns, tim *t, cfg *c, aap *aap, allocation *alloc, int rseq) {
    aap_aclm *aclm = malloc(sizeof(aap_aclm));
  
    aclm->n = n;
    aclm->ns = ns;
    aclm->t = t;
    aclm->cfg = c;
    aclm->aap = aap;
    aclm->alloc  = alloc;
    aclm->rseq = rseq;
    aclm->mseq = 0;
  
  
    aclm->send = send_msg;
    aclm->delete = delete;  
    aclm->range = range;
    aclm->set_allocation = set_allocation;
    return aclm;
}
