#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "aap_msg.h"

#if BYTE_ORDER == BIG_ENDIAN || defined(_BIG_ENDIAN)
#define htonl24(x) (x & 0xffffff)
#else
#define htonl24(x) (((x & 0xff) << 16) | ((x & 0xff00)) | ((x & 0xff0000) >> 16))
#endif

#define ntohl24(x) htonl24(x)

static void create(aap_msg *msg, char msgtype, short addrfamily, int rseq, char mseq,
		   char *data, int len);
static void aap_send(aap_msg *msg, net *n, net_socket *ns);
static int msgtype(aap_msg *msg);
static int addrfamily(aap_msg *msg);
static int rseq(aap_msg *msg);
static int mseq(aap_msg *msg);
static void body(aap_msg *msg, char **dataptr, int *len);
static void debug_print(aap_msg *msg);  /* Should be removed when released */

/*------------------------------------------------------------------*/
static void create(aap_msg *msg, char msgtype, short addrfamily, int rseq, char mseq,
		   char *data, int len) {
    int i;
    msg->data.version = 0;
    msg->data.msgtype = msgtype;
    msg->data.addrfamily = htons(addrfamily);
    msg->data.rseq = htonl24(rseq);
    msg->data.mseq = mseq;
    for(i = 0; i < len; i++) {
        msg->data.body[i] = data[i];
    }
    msg->bodylen = len;
}
/*------------------------------------------------------------------*/
static int parse(aap_msg *msg, char *data, int len, net_peer *np) {
    int i;

    /* Received AAP packets with a length less than 12
       should be discarded. The return value of -1 indicates
       just that. */
    if(len < 12) {
        printf("aap_msg: len < 12 (%d)\n", len);
        return -1;
    }

    for(i = 0; i < len; i++) {
        ((char *)(&(msg->data)))[i] = data[i];
    }
  
    /* for now, only version 0 packets are allowed */
    if(msg->data.version != 0) {
        printf("aap_msg: version != 0\n");
        return -1;
    }

    /* packets with a message type other than the allowed
       should be discarded. */
    if(msg->data.msgtype >= AAP_MSG_MAX) {
        printf("aap_msg: error in message type (%d)\n", msg->data.msgtype);
        return -1;
    }
    msg->sender_addr = np->peer_addr;

    msg->bodylen = len - 8;
    return 0;
}
/*------------------------------------------------------------------*/
static void aap_send(aap_msg *msg, net *n, net_socket *ns) {
    n->send(n, ns, (char *)(&(msg->data)), 8 + msg->bodylen);
}
/*------------------------------------------------------------------*/
static int msgtype(aap_msg *msg) {
    return msg->data.msgtype;
}
/*------------------------------------------------------------------*/
static int addrfamily(aap_msg *msg) {
    return ntohs(msg->data.addrfamily);
}
/*------------------------------------------------------------------*/
static int rseq(aap_msg *msg) {
    return ntohl24(msg->data.rseq);
}
/*------------------------------------------------------------------*/
static int mseq(aap_msg *msg) {
    return msg->data.mseq;
}
/*------------------------------------------------------------------*/
static unsigned long sender(aap_msg *msg) {
    return msg->sender_addr;
}
/*------------------------------------------------------------------*/
static void body(aap_msg *msg, char **dataptr, int *len) {
    *dataptr = (char *)&(msg->data.body);
    *len     = msg->bodylen;
}
/*------------------------------------------------------------------*/

static void debug_print(aap_msg *msg) {
    char *msgtype;
    struct in_addr in;

    struct hostent *he;

    /*  int i;
        for(i = 0; i < 8; i++) {
        printf("%02x ", ((char *)&(msg->data))[i]);    
        }
        printf("\n");*/
  
    in.s_addr = msg->sender_addr;
    he = gethostbyaddr((char *)&in, sizeof(in), AF_INET);

    switch(msg->data.msgtype) {
    case AAP_MSG_ACLM:
        msgtype = "ACLM";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    case AAP_MSG_AIU:
        msgtype = "AIU";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    case AAP_MSG_AITU:
        msgtype = "AITU";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    case AAP_MSG_ASA:
        msgtype = "ASA";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    case AAP_MSG_ASRP:
        msgtype = "ASRP";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    case AAP_MSG_ANA:
        msgtype = "ANA";
        if(DEBUG_ENABLED('d') == 0){
            printf("%s message: ", msgtype);
            printf("ver 0x%x, ", msg->data.version);
            printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
            printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
            printf("mseq 0x%x ", msg->data.mseq);
            if(he != NULL) {
                printf("from %s\n", he->h_name);
            } else {
                printf("from %s\n", inet_ntoa(in));
            }
        }
        break;
    default:
        msgtype = "???";
        printf("%s message: ", msgtype);
        printf("ver 0x%x, ", msg->data.version);
        printf("addr fam 0x%x, ", ntohs(msg->data.addrfamily));
        printf("rseq 0x%x, ", (int)ntohl24(msg->data.rseq));
        printf("mseq 0x%x ", msg->data.mseq);
        if(he != NULL) {
            printf("from %s\n", he->h_name);
        } else {
            printf("from %s\n", inet_ntoa(in));
        }
    }
    
    /*in.s_addr = msg->sender_addr;
    he = gethostbyaddr((char *)&in, sizeof(in), AF_INET);

    if(he != NULL) {
        printf("from %s\n", he->h_name);
    } else {
        printf("from %s\n", inet_ntoa(in));
        }*/
}

/*------------------------------------------------------------------*/

aap_msg *aap_msg_new() {
    aap_msg *am = calloc(1, sizeof(aap_msg));
    am->create = create;
    am->parse = parse;
    am->send = aap_send;
    am->msgtype = msgtype;
    am->addrfamily = addrfamily;
    am->rseq = rseq;
    am->mseq = mseq;
    am->sender = sender;
    am->body = body;
    am->debug_print = debug_print;
  
    return am;
}
