#ifndef __AAP_MSG_H__
#define __AAP_MSG_H__

/**
 * aap_msg.h
 *
 * The datastructure that repressents an AAP message.
 * The data fields are (taken from draft-ietf-malloc-aap-03.txt):
 *
 *   FIELD      OCTETS       DESCRIPTION
 *   -----      ------       -----------
 *
 *   version       1  Protocol version number (zero for this specification)
 *   msgtype       1  Message type (ACLM, AIU, etc.)
 *   addrfamily    2  Address family (IPv4, IPv6, etc.)
 *   rseq          3  Request sequence number
 *   mseq          1  Message sequence number
 *   body        var  Body (format varies by msgtype)
 *
 * All quantities > 8 bits are stored in network byte order.
 */

#include "net.h"
#include "net_socket.h"

/* The IP address family as defined by the IANA */
#define AAP_MSG_IP 1

typedef struct aap_msg_ aap_msg;

enum aap_msg_type {
    AAP_MSG_ACLM,
    AAP_MSG_AIU,
    AAP_MSG_AITU,
    AAP_MSG_ASA,
    AAP_MSG_ASRP,
    AAP_MSG_ANA,
    AAP_MSG_MAX /* dummy message that only indicates the highest allowed
                   message identifier number */
};

struct aap_msg_data {
    unsigned char version;
    unsigned char msgtype;
    unsigned short addrfamily;
    /*#if BYTE_ORDER == LITTLE_ENDIAN
      unsigned int   mseq:8,           
      rseq:24;
      #else*/
    unsigned int   rseq:24, 
        mseq:8; 
    /*#endif*/
    char body[500];
};

struct aap_msg_ {
    struct aap_msg_data data;
    int bodylen;   /* the length of the body of the AAP message */
    unsigned long sender_addr;
    void (* create)(aap_msg *msg, char msgtype, short addrfamily, int rseq, char mseq,
                    char *data, int len);
    int (* parse)(aap_msg *msg, char *data, int len, net_peer *np);
    void (* send)(aap_msg *msg, net *n, net_socket *ns);
  
    int (* msgtype)(aap_msg *msg);
    int (* addrfamily)(aap_msg *msg);
    int (* rseq)(aap_msg *msg);
    int (* mseq)(aap_msg *msg);
    unsigned long (* sender)(aap_msg *msg);
    void (* body)(aap_msg *msg, char **dataptr, int *len);
  
    void (* debug_print)(aap_msg *msg);  /* Should be removed when released */
};

aap_msg *aap_msg_new();

#endif
