/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains constructor and methods for a generic MADCAP message.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "madcap_msg.h"
#include "debug.h"

static char err_msg[256];

/* Declaration of functions. */
static int pub_get_multicasted(madcap_msg *object, int16 *bool);

/**
 * Public method to create a MADCAP message.
 *
 * @param object The empty object created with new().
 * @param msgtype Type of MADCAP message.
 * @param addrfamily Address family, for now only IPv4 (= 1).
 * @param xid Transaction ID.
 * @param options A pointer to the options field.
 * @param options_len The lenght of the options field.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_create(madcap_msg *object, madcap_msg_type msgtype,
                      uint16 addrfamily, uint32 xid, uint8 *options,
                      int options_len){
    int i;
    if(options_len > MAX_OPTIONS) {
        sprintf(err_msg, "madcap_msg: options_len = %d.", options_len);
        mc_error->notify(mc_error, MC_ERR_OPTIONSLEN, err_msg);
        return FAILURE;
    }
    object->data.version = MADCAP_VERSION;
    object->data.msgtype = msgtype;
    object->data.addrfamily = htons(addrfamily);
    object->data.xid = htonl(xid);
    for(i = 0;i < options_len;i++) {
        object->data.options[i] = options[i];
    }
    object->options_len = options_len;
    return SUCCESS;
}

/**
 * Parse incoming message and check syntax.
 * Checks message for length, version, valid messagetype, address family, and
 * syntax of options field.
 * A flag is set for each contained option. All flags can be obtain through the
 * method get_options_flag. If options field is incorrect the first incorrect
 * option is indicated by the same flag. If the incorrect option cannot be
 * identified all flags are set.
 *
 * @param object The object message to parse.
 * @param data A pointer to message data.
 * @param datalen Length of message data.
 * @param np The address of the sender.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_parse(madcap_msg *object, uint8 *data, int datalen,
                     net_peer *np){
       
    int i = 0;
    uint16 code, len;

    /* State variable for parsing of MADCAP message syntax.
       Set to FAILURE if message incorrect. */
    enum return_values msg_syntax = SUCCESS;

    /* Check minimum lenght of message. */
    if(datalen < MIN_MSG_LENGTH) {
        sprintf(err_msg, "madcap_msg: datalen = %d.", datalen); 
        mc_error->notify(mc_error, MC_ERR_MSGLEN, err_msg);
        return FAILURE;
    }
    
    /* Copy data. */
    for(i = 0;i < datalen;i++) {
        ((uint8 *) (&(object->data)))[i] = data[i];
    }
    
    /* Check for correct version. */
    if(object->data.version != MADCAP_VERSION) {
        sprintf(err_msg, "madcap_msg: version = %d.",
                object->data.version);
        mc_error->notify(mc_error, MC_ERR_VERSION, err_msg);
        return FAILURE;
    }
    
    /* Check for valid message type. */
    if(object->data.msgtype >= MC_MSGTYPE_VALID) {
        sprintf(err_msg, "madcap_msg: message type = %d.",
                object->data.msgtype);
        mc_error->notify(mc_error, MC_ERR_MSGTYPE, err_msg);
        return FAILURE;
    }
    
    /* Check if address family is supported. */
    if(ntohs(object->data.addrfamily) != MC_ADDR_FAMILY) {
        sprintf(err_msg, "madcap: address family = %d.",
                ntohs(object->data.addrfamily));
        mc_error->notify(mc_error, MC_ERR_FAMILY, err_msg);
        return FAILURE; 
    }  

    /* Converting xid from network to host byte order. */
    object->data.xid = ntohl(object->data.xid);
    
    
    /* Check options field syntax and option types. */
    /* Requirements as defined in RFC 2730:            Check:
     * -----------------------------------------       ------
     * For each option:                                       
     * - two octets optiontype                           x   
     *  (it is impossible to forsee the result of such an error)
     * - two octets optionlength                         x   
     *  (it is impossible to forsee the result of such an error)
     * - optionlength octets of option data              x    
     * End option last = four octets of zeroes.          x   
     * Only one option of each option type.              x   
     * Ignore unrecognized options and process anyway.   x   
     * If message does not fulfill requirements and is
     *  not a NAK, respond with a NAK and Error option.  in corresponding module
     */
    i = 0;
    msg_syntax = FAILURE;
    sprintf(err_msg,
            "madcap_msg: Option field lacks End option, option length of some\
 option wrong, or option type/length in only one octet.");
    while(i < (datalen - MADCAP_HEADER)) {
        code = object->data.options[i]; /* Network byte order - most
                                           significant byte first. */
        i++;
        code = code << 8;
        code = code | ((uint16) (object->data.options[i]));
        i++; /* Step to length value. */
        len = (uint16) object->data.options[i];
        i++;
        len = len << 8;
        len = len | (uint16) (object->data.options[i]);
        i++; /* Step to option value. */
        
        switch(code) {
        case MC_END:
            DEBUG('g', "Option 'End' identified.\n");
            i = MAX_OPTIONS; /* End for-loop. */
            if(len != 0){
                sprintf(err_msg, "Incorrect end of option field, option length\
 of some option wrong, or option type/length in only one octet.");
                object->incorrect = MC_OPT_VALID; /* Any option can be
                                                     incorrect. */
            } else {
                msg_syntax = SUCCESS;
            }
            break;
        case MC_LEASE_TIME:
            DEBUG('g', "Option 'Lease Time' identified.\n");
            if(((object->opt_types & 0x0001) == 0x0001) || (len != 4)){
                /* Option already occured once in message, or wrong length of
                   option data. */
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Lease Time' occurred\
 twice in message, or wrong length of option data.");
                object->incorrect = MC_LEASE_TIME; /* Flag to indicate which
                                                      option failed. */
                i = MAX_OPTIONS; /* End while-loop. */
            } else {
                /* Set flag for option. */
                object->opt_types = object->opt_types | 0x0001;
                /* Set pointer to option length. */
                object->options[MC_LEASE_TIME] = &(object->data.options[i-2]);
                i = i + len; /* Step to next option. */
            }
            break;
        case MC_SERVER_ID:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Server Identifier' identified.\n");
            if(((object->opt_types & 0x0002) == 0x0002) || (len < 3)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Server Identifier'\
 occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_SERVER_ID;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0002;
                object->options[MC_SERVER_ID] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_LEASE_ID:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Lease Identifier' identified.\n");
            if(((object->opt_types & 0x0004) == 0x0004) || (len < 1)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Lease Identifier'\
 occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_LEASE_ID;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0004;
                object->options[MC_LEASE_ID] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_SCOPE:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Multicast Scope' identified.\n");
            if(((object->opt_types & 0x0008) == 0x0008) || (len < 1)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Multicast Scope'\
 occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_SCOPE;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0008;
                object->options[MC_SCOPE] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_REQUEST_LIST:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Option Request List' identified.\n");
            if(((object->opt_types & 0x0010) == 0x0010) || (len < 2)) {
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Option Request List'\
 occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_REQUEST_LIST;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0010;
                object->options[MC_REQUEST_LIST] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_START_TIME:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Start Time' identified.\n");
            if(((object->opt_types & 0x0020) == 0x0020) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Start Time' occurred\
 twice in message, or wrong length of option data.");
                object->incorrect = MC_START_TIME;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0020;
                object->options[MC_START_TIME] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_NO_ADDR_REQ:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Number of Addresses Requested' identified.\n"); 
            if(((object->opt_types & 0x0040) == 0x0040) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Number of Addresses\
 Requested' occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_NO_ADDR_REQ;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0040;
                object->options[MC_NO_ADDR_REQ] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_REQ_LANGUAGE:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Requested Language' identified.\n");
            if((object->opt_types & 0x0080) == 0x0080){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Requested Language'\
 occurred twice in message.");
                object->incorrect = MC_REQ_LANGUAGE;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0080;
                object->options[MC_REQ_LANGUAGE] = &(object->data.options[i-2]);
                i = i + len;
            }
            break;
        case MC_SCOPE_LIST:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Multicast Scope List' identified.\n");
            if((object->opt_types & 0x0100) == 0x0100){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Multicast Scope List'\
 occurred twice in message.");
                object->incorrect = MC_SCOPE_LIST;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0100;
                object->options[MC_SCOPE_LIST] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_ADDR_RANGE_LIST:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'List of Address Ranges' identified.\n");
            if((object->opt_types & 0x0200) == 0x0200){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'List of Address Ranges'\
 occurred twice in message.");
                object->incorrect = MC_ADDR_RANGE_LIST;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0200;
                object->options[MC_ADDR_RANGE_LIST] =
                    &(object->data.options[i-2]);
                i = i + len;
            }            
            break;
        case MC_CURRENT_TIME:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Current Time' identified.\n");
            if(((object->opt_types & 0x0400) == 0x0400) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Current Time' occurred\
 twice in message, or wrong length of option data.");
                object->incorrect = MC_CURRENT_TIME;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0400;
                object->options[MC_CURRENT_TIME] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_FEATURE_LIST:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Feature List' identified.\n");
            if (((object->opt_types & 0x0800) == 0x0800) || (len < 6)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Feature List' occurred\
 twice in message, or wrong length of option data.");
                object->incorrect = MC_FEATURE_LIST;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x0800;
                object->options[MC_FEATURE_LIST] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_RETRY_TIME:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Retry Time' identified.\n");
            if(((object->opt_types & 0x1000) == 0x1000) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Retry Time' occurred\
 twice in message, or wrong length of option data.");
                object->incorrect = MC_RETRY_TIME;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x1000;
                object->options[MC_RETRY_TIME] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_MIN_LEASE_TIME:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Minimum Lease Time' identified.\n");
            if(((object->opt_types & 0x2000) == 0x2000) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Minimum Lease Time' occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_MIN_LEASE_TIME;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x2000;
                object->options[MC_MIN_LEASE_TIME] =
                    &(object->data.options[i-2]);
                i = i + len;
            }          
            break;
        case MC_MAX_START_TIME:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Maximum Start Time' identified.\n");
            if(((object->opt_types & 0x4000) == 0x4000) || (len != 4)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Maximum Start Time'\
 occurred twice in message, or wrong length of option data.");
                object->incorrect = MC_MAX_START_TIME;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x4000;
                object->options[MC_MAX_START_TIME] =
                    &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        case MC_ERROR:
            /* See above in 'case MC_LEASE_TIME' for comments. */
            DEBUG('g', "Option 'Error' identified.\n");
            if(((object->opt_types & 0x8000) == 0x8000) || (len < 2)){
                msg_syntax = FAILURE;
                sprintf(err_msg, "madcap_msg: option 'Error' occurred twice in\
 message, or wrong length of option data.");
                object->incorrect = MC_ERROR;
                i = MAX_OPTIONS;
            } else {
                object->opt_types = object->opt_types | 0x8000;
                object->options[MC_ERROR] = &(object->data.options[i-2]);
                i = i + len;
            }           
            break;
        default:
            /* Ignore unrecognized options. */
            DEBUG('g', "madcap_msg: Option type (code = %d) not recognized -\
 ignoring.\n", code);
            i = i + len;
            break;
        }
    }
    if(msg_syntax == FAILURE) {
        mc_error->notify(mc_error, MC_ERR_OPTIONS, err_msg);
        if(object->opt_types == 0) {
            object->incorrect = MC_OPT_VALID; /* Any option can be incorrect. */
        }
        DEBUG('g',"madcap_msg: Parse failed.\n");
        return FAILURE;
    }
    DEBUG('g',"Options field syntax correct.\n");
    
    
    /* Store sender and destination address of message. */
    if(np != NULL){
        object->sender_address = np->peer_addr;
        object->destination_address = np->dest_addr;
    }

    if(DEBUG_ENABLED('g') == 0){
        int16 multicast;
        pub_get_multicasted(object, &multicast);
        if(multicast == -1){
            printf("Incoming message multicasted.\n");
        } else {
            printf("Incoming message unicasted.\n");
        }
    }

    return SUCCESS;
} 



/**
 * Send a MADCAP message.
 *
 * @param object The message to send.
 * @param n The network abstraction.
 * @param ns The network connection abstraction.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_send(madcap_msg *object, net *n, net_socket *ns){
    if((n->send(n, ns, (uint8 *)(&(object->data)),
                 MADCAP_HEADER + object->options_len)) == FAILURE) {
        return FAILURE;
    }
    return SUCCESS;   
} 

/**
 * Delete object and free memory.
 *
 * @param object The object to delete.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_delete(madcap_msg *object){
    free(object);
    return SUCCESS;
}

/**
 * Method to get the message type of a MADCAP message.
 *
 * @param object The message object.
 * @param msgtype A pointer to store the message type.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_msg_type(madcap_msg *object, madcap_msg_type *msgtype){
    *msgtype = object->data.msgtype;
    return SUCCESS;    
}

/**
 * Method to get the address family of a MADCAP message.
 * For now only IPv4 is supported.
 *
 * @param object The message object.
 * @param addrfam A pointer to store the address family.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_addrfamily(madcap_msg *object, uint16 *addrfam){
    *addrfam = object->data.addrfamily;
    return SUCCESS;   
}

/**
 * Method to get the transaction ID of a MADCAP message.
 *
 * @param object The message object.
 * @param addrfam A pointer to store the transaction ID.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_xid(madcap_msg *object, uint32 *xid){
    *xid = object->data.xid;
    return SUCCESS;   
}

/**
 * Method to get the sender of a MADCAP message.
 *
 * @param object The message object.
 * @param sender A pointer to store the sender address.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_sender(madcap_msg *object, uint32 *sender){
    *sender = object->sender_address;
    return SUCCESS;   
}

/**
 * Method to get the destination address of a incoming MADCAP message.
 *
 * @param object The message object.
 * @param destination A pointer to store the destination address (network byte
 * order).
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_destination(madcap_msg *object, uint32 *destination){
    *destination = object->destination_address;
    return SUCCESS;
}

/**
 * Method to get if the incoming MADCAP message was multicasted.
 *
 * @param object The message object.
 * @param bool A pointer to store the boolean value, 0 if unicasted and -1 if
 * multicasted.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_multicasted(madcap_msg *object, int16 *bool){
    uint32 min, max, dest_addr;
    
    min = ntohl(inet_addr("224.0.0.0"));
    max = ntohl(inet_addr("239.255.255.255"));
    dest_addr = ntohl(object->destination_address);

    if(dest_addr >= min && dest_addr <= max) {
        *bool = -1;
    } else {
        *bool = 0;
    }

    return SUCCESS;
}


/**
 * Method to get which options was included in the incoming message
 * The option Lease Time is indicated by the least significant bit, and the
 * option Error is indicated by the most significant bit. Option End has no
 * flag.
 *
 * @param object The message object.
 * @param options A pointer to store the options.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_options_flag(madcap_msg *object, uint16 *options){
    *options = object->opt_types;
    return SUCCESS;   
}

/**
 * Method to get specific option in message.
 *
 * @param object The message object.
 * @param opt Desired option to get.
 * @param optlen The length of the option data.
 * @param optiondata A pointer to be set to desired option data, i.e the
 *        address to a uint8 pointer.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_option(madcap_msg *object, madcap_opt_type opt,
                          uint16 *optlen, uint8 **optiondata){
    uint16 flag = 1;
    flag = flag << (opt - 1);

    /* Check if requested option is valid and exists in message. */
    if((opt >= MC_OPT_VALID) || ((flag & object->opt_types) != flag)) {
        sprintf(err_msg, "%d is not a valid option type for this message.",
                opt);
        mc_error->notify(mc_error, MC_ERR_UNSPECIFIED, err_msg);
        return FAILURE;
    }

    /* Get length of option data. */
    *optlen = (uint16) *(object->options[opt]);
    *optlen = *optlen << 8;
    *optlen = *optlen | (uint16) *((object->options[opt]) + 1);

    /* Set pointer to first byte (most significant) of option data. */
    *optiondata = ((object->options[opt]) + 2);

    return SUCCESS;
}

/**
 * Method to get which option was incorrect in parsed message.
 * If some syntax error in message this function gives which option is
 * incorrect. If impossible to state which option is incorrect it gives
 * MC_OPT_VALID (17). A value of -1 indicates no error.
 *
 * @param object The message object.
 * @param opt The option type that failed.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_get_error_option(madcap_msg *object, madcap_opt_type *opt){
    *opt = object->incorrect;
    return SUCCESS;
}

/**
 * The constructor of a madcap-msg object.
 *
 * @return A pointer to a madcap_msg object.
 */
madcap_msg *madcap_msg_new(){
    madcap_msg *object = calloc(1, sizeof(madcap_msg));
    ASSERT(object != NULL, "Memory allocation failed.");
    /*TODO: Possible improvement would be if object->data.options is
      filled with specific data. An error in option length can easier be
      identified, as well as an option type with only one octet.*/
    object->create = pub_create;
    object->parse = pub_parse;
    object->send = pub_send;
    object->delete = pub_delete;
    object->get_msg_type = pub_get_msg_type;
    object->get_addrfamily = pub_get_addrfamily;
    object->get_xid = pub_get_xid;
    object->get_sender = pub_get_sender;
    object->get_destination = pub_get_destination;
    object->get_multicasted = pub_get_multicasted;
    object->get_options_flag = pub_get_options_flag;
    object->get_option = pub_get_option;
    object->get_error_option = pub_get_error_option;
    
    object->opt_types = 0;
    object->incorrect = -1;
    
    return object;
}
