/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains functions that is used when the server receives a
 * message it does not recognize.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */


#include "madcap_msg_unknown.h"
static int pub_receive(madcap_msg_unknown *object);
static int pub_delete(madcap_msg_unknown *object);
extern madcap_error *mc_error;

/**
 * Creates a new UNKNOWN message.
 * 
 * @param msg An incoming madcap_msg message of unknown type.
 * @param n A network abstraction object.
 * @param ns An abstract network connection.
 * @param error The type of error that causes this message.
 * @return The newly created message. The message should be deleted using the
 * delete function. A return value of NULL indicates a failure to create the
 * new message.
 */
madcap_msg_unknown *madcap_msg_unknown_new(madcap_msg *msg,
                                           net *n,
                                           net_socket *ns,
                                           madcap_error_type error) {

    madcap_msg_unknown *object;

    DEBUG('m',"Creating UNKNOWN message.\n");
    
    object = (madcap_msg_unknown *) malloc(sizeof(madcap_msg_unknown));
    ASSERT(object != NULL, "Memory allocation failed.");
    
    object->msg      = msg;
    msg->get_options_flag(msg, &(object->opt_types_incl));
    object->err_type = error;
    msg->get_error_option(msg, &(object->err_type_opt));
    object->n        = n;
    object->ns       = ns;
    msg->get_xid(msg, &(object->xid));
    object->receive  = pub_receive;
    object->delete   = pub_delete;

    return object;    
}


/*
 * Receives a UNKNOWN message.
 * The function takes the appropriate actions when a UNKNOWN message is
 * received. The function will send an NAK to the client as response.
 *
 * @param object The message (object) that is received.
 * @return 0 on success and -1 on failure.
 */
static int pub_receive(madcap_msg_unknown *object) {

    madcap_msg_nak *nak;
    madcap_msg_format_options *format;
    int send_nak = TRUE;
    int result = SUCCESS;
    int ok;                         /* Temp variable. */
    int check;                      /* Temp variable. */
    uint8 *opt_data;                /* Option data from incoming messages. */
    uint16 opt_data_len;            /* Length of opt_data. */

    uint8 *lease_id                = NULL; /* Used with option 3. */
    uint16 lease_id_len            = 0;    /* Used with option 3. */
    madcap_nak_error_code err_code = 0;    /* Used with option 16. */
    uint16 err_option              = 0;    /* Used with option 16. */
    int err_option_count           = 0;    /* Used with option 16. */
    int err_data_len               = 2;    /* Used with option 16. */

    DEBUG('m',"Receiving UNKNOWN message.\n");

    /* Check if required option LEASE ID (3) is present and valid. */
    ok = (object->opt_types_incl & 0x0004) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_LEASE_ID,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len > 0) && (opt_data != NULL)) {
            lease_id     = opt_data;
            lease_id_len = opt_data_len;
            DEBUG('m',"Lease ID present but not valid. NAK is not sent.\n");
        }
    } else {
        send_nak = FALSE;
        DEBUG('m',"Lease ID not present. NAK is not sent.\n");
    }

    if(send_nak == TRUE) {
        /* Formating error code and error option. */
        err_code           = MC_NAK_INVALID_REQ;
        if(object->err_type == MC_ERR_MSGTYPE) {
            err_option     = 0xffff;
        } else if(object->err_type == MC_ERR_OPTIONS) {
            if((object->err_type_opt >= 0) |
               (object->err_type_opt < MC_OPT_VALID)) {
                err_option = htons(object->err_type_opt);
            } else {
                err_option = 0xffff;
                err_option_count++;
            }
        } else {
            err_option     = 0xffff;
            err_option_count++;
        }

        /* Create NAK message. */
        object->data     = NULL;
        object->data_len = 0;
        format           = madcap_msg_format_options_new();
        ASSERT(format != NULL, "Failed to create object.");
        ok               = SUCCESS;
        
        /* Add the SERVER IDENTIFIER (2) option (NAK). */
        ok = ok + format->server_id(format,
                                    &(object->data),
                                    &(object->data_len));
        
        /* Add the LEASE IDENTIFIER (3) option (NAK). */
        if(lease_id_len > 0) {
            ok = ok + format->lease_id(format,
                                       lease_id,
                                       lease_id_len,
                                       &(object->data),
                                       &(object->data_len));
        }
        
        /* Add the FEATURE LIST (12) option (NAK). */
        /* TODO: */
        
        /* Add the ERROR (16) option (NAK). */
        if(err_option_count != 1) {
            /* Problem was not with a specific option. */
            err_option = 0xffff;
        }
        err_option = htons(err_option);
        ok = ok + format->error(format,
                                err_code,
                                (uint8 *) &err_option,
                                err_data_len,
                                &(object->data),
                                &(object->data_len));
        
        /* Add the END (0) option (NAK). */
        ok = ok + format->end(format,
                              &(object->data),
                              &(object->data_len));
        
        /* Send message if options were added correctly. */
        if(ok == SUCCESS) {
            /* Send NAK. */
            nak = madcap_msg_nak_new(object->n,
                                     object->ns,
                                     object->xid,
                                     object->data,
                                     object->data_len);
            if(nak != NULL) {
                ok = nak->send(nak);
                nak->delete(nak);
            } else {
                ok = FAILURE;
            }
            result = ok;
        }

        /* Create error message if something failed. */
        if(ok == FAILURE) {
            mc_error->notify(mc_error, MC_ERR_CREATE_NAK,
                             "madcap_msg_unknown: Failed to create and send.");
            result = FAILURE;
        }
        
        format->delete(format);
    }

    return result;
}


/*
 * Deletes a UNKNOWN message.
 * The function deletes a UNKNOWN message and returnes used resources to the
 * system.
 *
 * @param object The message (object) that should be deleted.
 * @return 0 on success and -1 on failure.
 */
static int pub_delete(madcap_msg_unknown *object) {

    free(object->data);
    free(object);

    return SUCCESS;
}


