
#include <stdlib.h>
#include "madcap_plist.h"

/**
 * madcap_plist.c
 *
 * Implements a priority queue. See madcap_plist.h for details.
 */

#define FIRST(l) (l)->head

typedef struct madcap_plist_node node;


static int insert(madcap_plist *l, uint32 priority, void *item);
static void *first(madcap_plist *l);
static uint32 priority(madcap_plist *l, int pos);
static int pub_remove(madcap_plist *l, int pos);
static int is_empty(madcap_plist *l);
static void *find_elem_with_prio(madcap_plist *p, uint32 priority);
static int length(madcap_plist *p);
static int find(madcap_plist *p, void *item);
static void *elem_at(madcap_plist *p, int pos);		 

static void delete(madcap_plist *l);

/*------------------------------------------------------------------*/
static node *node_at(madcap_plist *l, int pos) {
  node *p;
  int i;
  if(pos > l->last_pos) {
    p = l->last_node;
    for(i = l->last_pos; i < l->len; i++) {
      if(i == pos) {
	l->last_pos = i;
	l->last_node = p;  
	return p;
      }
      p = p->next;
    }    
  } else {
    p = FIRST(l);
    for(i = 0; i < l->len; i++) {
      if(i == pos) {
	l->last_pos = i;
	l->last_node = p;
	return p;
      }
      p = p->next;
    }
  }
  return NULL;
}

/*------------------------------------------------------------------*/
static int insert(madcap_plist *l, uint32 priority, void *item) {
  node *p = FIRST(l);  
  node *q = NULL;     
  node *this;
  
  if((this = calloc(1, sizeof(node))) == NULL) {
    return -1;
  }
  
  this->priority = priority;
  this->item     = item;

  if(is_empty(l)) {
    this->next = NULL;
    l->len = 1;
    FIRST(l) = this;
    l->last_pos = 0;
    l->last_node = FIRST(l);  
    return 0;
  }
  while(p != NULL && p->priority < this->priority) {
    q = p;
    p = p->next;    
  }

  this->next = p;
  if(q != NULL) {
    q->next = this;    
  } else { /* q == NULL, this is the first node */
    FIRST(l) = this;
  }
  l->len++;
  l->last_pos = 0;
  l->last_node = FIRST(l);
  
  return 0;
}

/*------------------------------------------------------------------*/
/* first returns a pointer to the item of the first node in the queue */  
static void *first(madcap_plist *l) {
  return elem_at(l, 0);
}

/*------------------------------------------------------------------*/
/* priority returns the priority of the item at position pos.
 returns 0 when failing.
*/
static uint32 priority(madcap_plist *l, int pos) {
  if(is_empty(l)) {
    return 0;
  }  
  return node_at(l, pos)->priority;    
}

/*------------------------------------------------------------------*/
static int pub_remove(madcap_plist *l, int pos) {
  node *p, *q;
  int i;
  
  if(is_empty(l)) {
    return -1;
  }

  p = FIRST(l);
  q = NULL;
  i = 0;
  
  while(p != NULL) {
    if(i == pos) {
      if(q != NULL) {
	q->next = p->next;
	free(p);
      } else {
	/* p is first element in list */
	FIRST(l) = p->next;
	free(p);
      }
      l->len--;
      l->last_pos = 0;
      l->last_node = FIRST(l);  
      return 0;
    } else {
      q = p;
      p = p->next;
      i++;
    }
  }
  l->last_pos = 0;
  l->last_node = FIRST(l); 
  return -1;
}

/*------------------------------------------------------------------*/
static void remove_element(madcap_plist *l, int (* matching_function)(void *obj, void *element), void *obj) {
  node *q, *p;

  p = FIRST(l);
  q = NULL;
  
  while(p != NULL) {
    if(matching_function(obj, p->item) != 0) {
      if(q != NULL) {
	q->next = p->next;
	free(p);
	p = q->next;	
      } else {
	/* p is first element in list */
	FIRST(l) = p->next;
	free(p);
	p = FIRST(l);
      }
      l->len--;
    } else {
      q = p;
      p = p->next;
    }
  }
  l->last_pos = 0;
  l->last_node = FIRST(l);  

}

static void *find_elem_with_prio(madcap_plist *l, uint32 priority){
    node *p;
    
    if(is_empty(l)) {
        return NULL;
    }
    
    p = FIRST(l);
/*    q = NULL; */
    
    while(p != NULL){
        if(p->priority == priority){
            return p->item;
        }else{
/*            q = p; */
            p = p->next;
        }
    }
    return NULL;    
}

/*------------------------------------------------------------------*/
static void *find_element(madcap_plist *l, int (* matching_function)(void *obj, void *element), void *obj) {
    node *q, *p;
    
    p = FIRST(l);
    q = NULL;
    
    while(p != NULL) {
        if(matching_function(obj, p->item) != 0) {  /* found element */            
            return p->item; /* return pointer to object in plist */
        } else {
            q = p;
            p = p->next;
        }
    }
    return NULL;
}
/*------------------------------------------------------------------*/
static int length(madcap_plist *p) {
    return p->len;
}
/*------------------------------------------------------------------*/
static int find(madcap_plist *l, void *item) {
  node *p;
  int i;

  p = FIRST(l);
  for(i = 0; i < l->len; i++) {
    if(p->item == item) {
      return i;
    }
    p = p->next;
  }
  return -1;
}
/*------------------------------------------------------------------*/
static void *elem_at(madcap_plist *l, int pos) {
  node *p;
  p = node_at(l, pos);
  if(p != NULL) {
    return p->item;
  } else {
    return NULL;
  }
}

/*------------------------------------------------------------------*/
/* is_empty tests if the queue is empty.
 */
static int is_empty(madcap_plist *l) {  
  return FIRST(l) == NULL;
}
/*------------------------------------------------------------------*/
static void delete(madcap_plist *l) {
  node *q, *p = FIRST(l);
  while(p != NULL) {
    q = p->next;
    free(p);
    p = q;
  }
  free(l);
}
/*------------------------------------------------------------------*/
madcap_plist *madcap_plist_new() {
  madcap_plist *l;
  
  if((l = calloc(1, sizeof(madcap_plist))) == NULL) {
    return NULL;
  }

  l->len = 0;
  FIRST(l) = NULL;

  l->last_pos = 0;
  l->last_node = FIRST(l);
  
  l->insert   = insert;
  l->first    = first;
  l->priority = priority;
  l->remove   = pub_remove;
  l->remove_element = remove_element;
  l->find_element = find_element;
  l->find_elem_with_prio =find_elem_with_prio;
  
  l->length = length;
  l->find   = find;
  l->elem_at = elem_at;
  
  l->is_empty = is_empty;
  l->delete   = delete;
  return l;
  
}
